<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>EyeCatching</title>
</head>
<body>


    <form action="" enctype='multipart/form-data' method='POST'>
        <input type="file" name="file" style="margin-top:20px;"/>
        <br>
        <i>Input Figure Name</i>
        <input type="text" class="txt_input" name="name"  value="EyeCatching.jpeg" style="margin-top:10px;"/>
        <i>Points Per Plot</i>
        <input type="number" name="plot_points" min="1" max="10" value="5" style="margin-top:10px;"/>
    <input type="submit" value="submit" class="button-new" style="margin-top:15px;"/>
    </form>


    <img src="{{ url_for('static', filename= './'+userinput,_t=val1) }}"  alt="Your Figure is missing"/>

    <img src="{{ url_for('video_feed') }}">

</body>
</html>